#define _GNU_SOURCE
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <signal.h>
#include <malloc.h>
#include <rfb/rfb.h>
#include <xs.h>
#include "xenfb.h"

static void on_kbd_event(rfbBool down, rfbKeySym keycode, rfbClientPtr cl)
{
	extern uint32_t gdk_linux_mapping[0x10000];
	rfbScreenInfoPtr server = cl->screen;
	struct xenfb *xenfb = server->screenData;
	xenfb_send_key(xenfb, down, gdk_linux_mapping[keycode & 0xFFFF]);
}

static void on_ptr_event(int buttonMask, int x, int y, rfbClientPtr cl)
{
	static int last_x = -1, last_y = -1;
	static int last_button = -1;
	rfbScreenInfoPtr server = cl->screen;
	struct xenfb *xenfb = server->screenData;

	if (last_button != -1) {
		int i;

		for (i = 0; i < 8; i++) {
			if ((last_button & (1 << i)) !=
			    (buttonMask & (1 << i))) {
				printf("%d %d\n", buttonMask & (1 << i), i);
				xenfb_send_button(xenfb, buttonMask & (1 << i),
						  2 - i);
			}
		}
	}

	if (last_x != -1)
		xenfb_send_motion(xenfb, x - last_x, y - last_y);

	last_button = buttonMask;

	last_x = x;
	last_y = y;
}

static void xenstore_write_vncport(struct xs_handle *xsh, int port, int domid)
{
    char buf[256], *path;
    char *portstr = NULL;

    path = xs_get_domain_path(xsh, domid);
    if (path == NULL) {
        fprintf(stderr, "xs_get_domain_path() error\n");
        goto out;
    }

    if (snprintf(buf, 256, "%s/console/vnc-port", path) == -1)
	goto out;

    portstr = malloc(10);
    if (snprintf(portstr, 10, "%d", port) == -1)
	goto out;

    if (xs_write(xsh, XBT_NULL, buf, portstr, strlen(portstr)) == 0)
        fprintf(stderr, "xs_write() vncport failed\n");

 out:
    free(portstr);
}

static int xenstore_read_vncpasswd(struct xs_handle *xsh, int domid, char *pwbuf, int pwbuflen)
{
    char buf[256], *path, *uuid = NULL, *passwd = NULL;
    unsigned int i, len, rc = 0;

    if (xsh == NULL) {
	return -1;
    }

    path = xs_get_domain_path(xsh, domid);
    if (path == NULL) {
        fprintf(stderr, "xs_get_domain_path() error\n");
        return -1;
    }

    snprintf(buf, 256, "%s/vm", path);
    uuid = xs_read(xsh, XBT_NULL, buf, &len);
    if (uuid == NULL) {
        fprintf(stderr, "xs_read(): uuid get error\n");
	free(path);
	return -1;
    }

    snprintf(buf, 256, "%s/vncpasswd", uuid);
    passwd = xs_read(xsh, XBT_NULL, buf, &len);
    if (passwd == NULL) {
        free(uuid);
	free(path);
	return rc;
    }

    strncpy(pwbuf, passwd, pwbuflen-1);
    pwbuf[pwbuflen-1] = '\0';

    fprintf(stderr, "VNC password read from XS '%s'\n", pwbuf);

    passwd[0] = '\0';
    snprintf(buf, 256, "%s/vncpasswd", uuid);
    if (xs_write(xsh, XBT_NULL, buf, passwd, len) == 0) {
        fprintf(stderr, "xs_write() vncpasswd failed\n");
	rc = -1;
    }

    free(passwd);
    free(uuid);
    free(path);

    return rc;
}


static void vnc_update(struct xenfb *xenfb, int x, int y, int w, int h)
{
	rfbScreenInfoPtr server = xenfb->user_data;
	rfbMarkRectAsModified(server, x, y, x + w, y + h);
}

static struct option options[] = {
    { "domid", 1, NULL, 'd' },
    { "vncport", 1, NULL, 'p' },
    { "title", 1, NULL, 't' },
    { "unused", 0, NULL, 'u' },
    { "listen", 1, NULL, 'l' },
};

int main(int argc, char **argv)
{
	rfbScreenInfoPtr server;
	char *fake_argv[7] = { "vncfb", "-rfbport", "5901",
                               "-desktop", "xen-vncfb",
                               "-listen", "0.0.0.0" };
	int fake_argc = sizeof(fake_argv) / sizeof(fake_argv[0]);
	int domid = -1, port = -1;
        char * title = NULL;
        char * listen = NULL;
	struct xenfb *xenfb;
	fd_set readfds;
	int fd;
	char buffer[1024];
        int opt;
        bool unused = FALSE;
	struct xs_handle *xsh;
	char vncpasswd[1024];

	vncpasswd[0] = '\0';

	while ((opt = getopt_long(argc, argv, "d:p:t:u", options,
				  NULL)) != -1) {
		switch (opt) {
                case 'd':
                    domid = strtol(optarg, NULL, 10);
                    break;
                case 'p':
                    port = strtol(optarg, NULL, 10);
                    break;
                case 't':
                    title = strdup(optarg);
                    break;
                case 'u':
                    unused = TRUE;
                    break;
                case 'l':
                    listen = strdup(optarg);
                    break;
                }
        }
        if (optind != argc) {
            fprintf(stderr, "Invalid options!\n");
            exit(1);
        }
        if (domid == -1) {
            fprintf(stderr, "Domain ID must be specified!\n");
            exit(1);
        }

        if (port == -1)
            port = 5900 + domid;
	snprintf(buffer, sizeof(buffer), "%d", port);
	fake_argv[2] = buffer;

        if (title != NULL)
            fake_argv[4] = title;

        if (listen != NULL)
            fake_argv[6] = listen;

	signal(SIGPIPE, SIG_IGN);

	xenfb = xenfb_new();
	if (xenfb == NULL) {
		fprintf(stderr, "Could not create framebuffer (%s)\n",
			strerror(errno));
		exit(1);
	}

	if (!xenfb_attach_dom(xenfb, domid)) {
		fprintf(stderr, "Could not connect to domain (%s)\n",
			strerror(errno));
		exit(1);
	}

	xsh = xs_daemon_open();
	if (xsh == NULL) {
	        fprintf(stderr, "cannot open connection to xenstore\n");
		exit(1);
	}


	if (xenstore_read_vncpasswd(xsh, domid, vncpasswd, sizeof(vncpasswd)/sizeof(char)) < 0) {
		fprintf(stderr, "cannot read VNC password from xenstore\n");
		exit(1);
	}
	  

	server = rfbGetScreen(&fake_argc, fake_argv, 
			      xenfb->width, xenfb->height,
			      8, 3, xenfb->depth / 8);
	if (server == NULL) {
		fprintf(stderr, "Could not create VNC server\n");
		exit(1);
	}

	xenfb->user_data = server;
	xenfb->update = vnc_update;

        if (unused)
            server->autoPort = TRUE;

	if (vncpasswd[0]) {
	    char **passwds = malloc(sizeof(char**)*2);
	    if (!passwds) {
	        fprintf(stderr, "cannot allocate memory (%s)\n", strerror(errno));
		exit(1);
	    }
	    fprintf(stderr, "Activated password '%s'\n", vncpasswd);
            passwds[0] = vncpasswd;
            passwds[1] = NULL;

	    server->authPasswdData = passwds;
	    server->passwordCheck = rfbCheckPasswordByList;
	} else {
	  fprintf(stderr, "Running with no password\n");
	}
	server->serverFormat.redShift = 16;
	server->serverFormat.greenShift = 8;
	server->serverFormat.blueShift = 0;
	server->kbdAddEvent = on_kbd_event;
	server->ptrAddEvent = on_ptr_event;
	server->frameBuffer = (char *)xenfb->pixels;
	server->screenData = xenfb;
	server->cursor = NULL;
	rfbInitServer(server);

	rfbRunEventLoop(server, -1, TRUE);

	fd = xenfb_get_fileno(xenfb);

	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);

        xenstore_write_vncport(xsh, server->port, domid);

	while (select(fd + 1, &readfds, NULL, NULL, NULL) != -1) {
		if (FD_ISSET(fd, &readfds)) {
			xenfb_on_incoming(xenfb);
		}
		
		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
	}

	rfbScreenCleanup(server);
	xenfb_delete(xenfb);

	return 0;
}
