#include <malloc.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <xenctrl.h>
#include <linux/xenfb.h>
#include <linux/xenkbd.h>
#include <sys/select.h>
#include <stdbool.h>
#include <xen/linux/evtchn.h>
#include <xen/event_channel.h>
#include <sys/mman.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <xs.h>

#include "xenfb.h"

// FIXME defend against malicous backend?

struct xenfb_private
{
	struct xenfb pub;
	int domid;
	unsigned long fbdev_mfn, kbd_mfn;
	int fbdev_evtchn, kbd_evtchn;
	evtchn_port_t fbdev_port, kbd_port;
	int evt_xch;
	int xc;
	unsigned char *fb;
	struct xenfb_page *fb_info;
	struct xenkbd_info *kbd_info;
	unsigned long *fbmfns;
	int n_fbmfns, n_fbdirs;
};

struct xenfb *xenfb_new(void)
{
	struct xenfb_private *xenfb = malloc(sizeof(*xenfb));

	if (xenfb == NULL)
		return NULL;

	memset(xenfb, 0, sizeof(*xenfb));

	xenfb->domid = -1;

	xenfb->evt_xch = xc_evtchn_open();
	if (xenfb->evt_xch == -1) {
		int serrno = errno;
		free(xenfb);
		errno = serrno;
		return NULL;
	}

	xenfb->xc = xc_interface_open();
	if (xenfb->xc == -1) {
		int serrno = errno;
		xc_evtchn_close(xenfb->evt_xch);
		free(xenfb);
		errno = serrno;
		return NULL;
	}

	return &xenfb->pub;
}

int xenfb_get_fileno(struct xenfb *xenfb_pub)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;

	return xc_evtchn_fd(xenfb->evt_xch);
}

static void xenfb_detach_dom(struct xenfb_private *xenfb)
{
	xenfb->domid = -1;
	munmap(xenfb->fb, xenfb->fb_info->mem_length);
	munmap(xenfb->fb_info, XC_PAGE_SIZE);
	munmap(xenfb->kbd_info, XC_PAGE_SIZE);
	xc_evtchn_unbind(xenfb->evt_xch, xenfb->fbdev_port);
	xc_evtchn_unbind(xenfb->evt_xch, xenfb->kbd_port);
}

void xenfb_delete(struct xenfb *xenfb_pub)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	if (xenfb->domid != -1)
		xenfb_detach_dom(xenfb);
	free(xenfb);
}

static int xenfb_fb_event(struct xenfb_private *xenfb, union xenfb_in_event *event)
{
	uint32_t prod;
	struct xenfb_page *info = xenfb->fb_info;

	prod = info->in_prod;
	if (prod - info->in_cons == XENFB_IN_RING_LEN) {
	    errno = EAGAIN;
	    return -1;
	}

	mb();			/* ensure ring space available */
	XENFB_IN_RING_REF(info, prod) = *event;
	wmb();			/* ensure ring contents visible */
	info->in_prod = prod + 1;
	return xc_evtchn_notify(xenfb->evt_xch, xenfb->fbdev_port);
}

static int xenfb_kbd_event(struct xenfb_private *xenfb, union xenkbd_in_event *event)
{
	uint32_t prod;
	struct xenkbd_info *info = xenfb->kbd_info;

	prod = info->in_prod;
	if (prod - info->in_cons == XENKBD_IN_RING_LEN) {
	    errno = EAGAIN;
	    return -1;
	}

	mb();			/* ensure ring space available */
	XENKBD_IN_RING_REF(info, prod) = *event;
	wmb();			/* ensure ring contents visible */
	info->in_prod = prod + 1;
	return xc_evtchn_notify(xenfb->evt_xch, xenfb->kbd_port);
}

static char *xenfb_path_in_dom(struct xs_handle *h,
			 unsigned domid, const char *path,
			 char *buffer, size_t size)
{
	char *domp = xs_get_domain_path(h, domid);
	int n = snprintf(buffer, size, "%s/%s", domp, path);
	free(domp);
	if (n >= size)
		return NULL;
	return buffer;
}

static int xenfb_xs_scanf1(struct xs_handle *xsh, unsigned domid,
			   const char *path, const char *fmt,
			   void *dest)
{
	char buffer[1024];
	char *p;
	int ret;

	p = xenfb_path_in_dom(xsh, domid, path, buffer, sizeof(buffer));
	p = xs_read(xsh, XBT_NULL, p, NULL);
	if (!p)
		return -ENOENT;
	ret = sscanf(p, fmt, dest);
	free(p);
	if (ret != 1)
		return -EDOM;
	return 0;
}

bool xenfb_attach_dom(struct xenfb *xenfb_pub, int domid)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	char buffer[1024];
	struct xs_handle *xsh;
	unsigned dummy;
	int ret;
	char *p, **vec;
	union xenfb_in_event event;

	if (xenfb->domid != -1) {
		xenfb_detach_dom(xenfb);
		if (domid == -1)
			return true;
	}

	xsh = xs_daemon_open_readonly();
	if (!xsh)
	    goto error;

	p = xenfb_path_in_dom(xsh, domid, "vfb", buffer, sizeof(buffer));
	if (!xs_watch(xsh, p, ""))
		goto error;
	p = xenfb_path_in_dom(xsh, domid, "vkbd", buffer, sizeof(buffer));
	if (!xs_watch(xsh, p, ""))
		goto error;

	for (;;) {
		ret = xenfb_xs_scanf1(xsh, domid, "vfb/page-ref", "%lu",
				      &xenfb->fbdev_mfn);
		if (ret == -ENOENT || ret == -EAGAIN)
			goto wait;
		if (ret < 0)
			goto error;
		ret = xenfb_xs_scanf1(xsh, domid, "vfb/event-channel", "%u",
				      &xenfb->fbdev_evtchn);
		if (ret == -ENOENT || ret == -EAGAIN)
			goto wait;
		if (ret < 0)
			goto error;
		ret = xenfb_xs_scanf1(xsh, domid, "vkbd/page-ref", "%lu",
				      &xenfb->kbd_mfn);
		if (ret == -ENOENT || ret == -EAGAIN)
			goto wait;
		if (ret < 0)
			goto error;
		ret = xenfb_xs_scanf1(xsh, domid, "vkbd/event-channel", "%u",
				      &xenfb->kbd_evtchn);
		if (ret == -ENOENT || ret == -EAGAIN)
			goto wait;
		if (ret < 0)
			goto error;
		break;

	wait:
		printf("Waiting...\n");
		vec = xs_read_watch(xsh, &dummy);
		if (!vec)
			goto error;
		free(vec);
	}
	xs_daemon_close(xsh);
	xsh = NULL;

	xenfb->fbdev_port = xc_evtchn_bind_interdomain(xenfb->evt_xch, domid,
						       xenfb->fbdev_evtchn);
	if (xenfb->fbdev_port == -1)
		goto error;

	xenfb->kbd_port = xc_evtchn_bind_interdomain(xenfb->evt_xch, domid,
						     xenfb->kbd_evtchn);
	if (xenfb->kbd_port == -1) 
		goto error_fbdev;

	xenfb->fb_info = xc_map_foreign_range(xenfb->xc, domid, XC_PAGE_SIZE,
					      PROT_READ | PROT_WRITE,
					      xenfb->fbdev_mfn);
	if (xenfb->fb_info == NULL)
		goto error_kbd;

	xenfb->kbd_info = xc_map_foreign_range(xenfb->xc, domid, XC_PAGE_SIZE,
					       PROT_READ | PROT_WRITE,
					       xenfb->kbd_mfn);
	if (xenfb->kbd_info == NULL)
		goto error_fbinfo;

	xenfb->n_fbmfns = (xenfb->fb_info->mem_length + (XC_PAGE_SIZE - 1)) / XC_PAGE_SIZE;
	xenfb->n_fbdirs = xenfb->n_fbmfns * sizeof(unsigned long);
	xenfb->n_fbdirs = (xenfb->n_fbdirs + (XC_PAGE_SIZE - 1)) / XC_PAGE_SIZE;

	xenfb->fbmfns = xc_map_foreign_batch(xenfb->xc, domid, PROT_READ, xenfb->fb_info->pd, xenfb->n_fbdirs);
	if (xenfb->fbmfns == NULL)
		goto error_kbdinfo;

	xenfb->fb = xc_map_foreign_batch(xenfb->xc, domid, PROT_READ | PROT_WRITE, xenfb->fbmfns, xenfb->n_fbmfns);
	if (xenfb->fb == NULL)
		goto error_fbmfns;

	event.type = XENFB_TYPE_SET_EVENTS;
	event.set_events.flags = XENFB_FLAG_UPDATE;
	if (xenfb_fb_event(xenfb, &event))
		goto error_fb;

	munmap(xenfb->fbmfns, xenfb->n_fbdirs * XC_PAGE_SIZE);

	xenfb->domid = domid;

	xenfb->pub.pixels = xenfb->fb;

	xenfb->pub.row_stride = xenfb->fb_info->line_length;
	xenfb->pub.depth = xenfb->fb_info->depth;
	xenfb->pub.width = xenfb->fb_info->width;
	xenfb->pub.height = xenfb->fb_info->height;

	return true;

 error_fb:
	printf("%d\n", __LINE__);
	{
		int serrno = errno;
		munmap(xenfb->fb, xenfb->fb_info->mem_length);
		errno = serrno;
	}
 error_fbmfns:
	printf("%d\n", __LINE__);
	{
		int serrno = errno;
		munmap(xenfb->fbmfns, xenfb->n_fbdirs * XC_PAGE_SIZE);
		errno = serrno;
	}
 error_kbdinfo:
	printf("%d\n", __LINE__);
	{
		int serrno = errno;
		munmap(xenfb->kbd_info, XC_PAGE_SIZE);
		errno = serrno;
	}
 error_fbinfo:
	printf("%d\n", __LINE__);
	{
		int serrno = errno;
		munmap(xenfb->fb_info, XC_PAGE_SIZE);
		errno = serrno;
	}
 error_kbd:
	printf("%d\n", __LINE__);
	{
		int serrno = errno;
		xc_evtchn_unbind(xenfb->evt_xch, xenfb->kbd_port);
		errno = serrno;
	}
 error_fbdev:
	printf("%d\n", __LINE__);
	{
		int serrno = errno;
		xc_evtchn_unbind(xenfb->evt_xch, xenfb->fbdev_port);
		errno = serrno;
	}
 error:
	printf("%d\n", __LINE__);
	if (xsh) {
		int serrno = errno;
		xs_daemon_close(xsh);
		errno = serrno;
	}

	return false;
}

static void xenfb_update(struct xenfb_private *xenfb, int x, int y, int width, int height)
{
	if (xenfb->pub.update)
		xenfb->pub.update(&xenfb->pub, x, y, width, height);
}

static void xenfb_on_fb_event(struct xenfb_private *xenfb)
{
	uint32_t prod, cons;
	struct xenfb_page *info = xenfb->fb_info;

	prod = info->out_prod;
	rmb();			/* ensure we see ring contents up to prod */
	for (cons = info->out_cons; cons != prod; cons++) {
		union xenfb_out_event *event = &XENFB_OUT_RING_REF(info, cons);

		switch (event->type) {
		case XENFB_TYPE_UPDATE:
			xenfb_update(xenfb, event->update.x, event->update.y, event->update.width, event->update.height);
			break;
		}
	}
	mb();			/* ensure we're done with ring contents */
	info->out_cons = cons;
	// FIXME need to notify?
}

static void xenfb_on_kbd_event(struct xenfb_private *xenfb)
{
	uint32_t prod, cons;
	struct xenkbd_info *info = xenfb->kbd_info;

	prod = info->out_prod;
	rmb();			/* ensure we see ring contents up to prod */
	for (cons = info->out_cons; cons != prod; cons++) {
		union xenkbd_out_event *event = &XENKBD_OUT_RING_REF(info, cons);

		switch (event->type) {
		default:
			break;
		}
	}
	mb();			/* ensure we're done with ring contents */
	info->out_cons = cons;
	// FIXME need to notify?
}

int xenfb_on_incoming(struct xenfb *xenfb_pub)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	evtchn_port_t port;

	port = xc_evtchn_pending(xenfb->evt_xch);
	if (port == -1)
		return -1;

	if (port == xenfb->fbdev_port) {
		xenfb_on_fb_event(xenfb);
	} else if (port == xenfb->kbd_port) {
		xenfb_on_kbd_event(xenfb);
	}

	if (xc_evtchn_unmask(xenfb->evt_xch, port) == -1)
		return -1;

	return 0;
}

int xenfb_send_key(struct xenfb *xenfb_pub, bool down, int keycode)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	union xenkbd_in_event event;

	event.type = XENKBD_TYPE_KEY;
	event.key.pressed = down ? 1 : 0;
	event.key.keycode = keycode;

	return xenfb_kbd_event(xenfb, &event);
}

int xenfb_send_motion(struct xenfb *xenfb_pub, int rel_x, int rel_y)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	union xenkbd_in_event event;

	event.type = XENKBD_TYPE_MOTION;
	event.motion.rel_x = rel_x;
	event.motion.rel_y = rel_y;

	return xenfb_kbd_event(xenfb, &event);
}

int xenfb_send_button(struct xenfb *xenfb_pub, bool down, int button)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	union xenkbd_in_event event;

	event.type = XENKBD_TYPE_BUTTON;
	event.button.pressed = down ? 1 : 0;
	event.button.button = button;

	return xenfb_kbd_event(xenfb, &event);
}
